#ifndef __DDWIN_H___
#define __DDWIN_H___

#include <windows.h>
#include <ddraw.h>
#include <assert.h>

typedef u_long u_int32_t;
typedef u_short u_int16_t;

class DirectDrawWin {
    static int count_bits(int v) {
	int j = 32;
	int i = 0;
	while (j--) {
	    if (v & 1) i ++;
	    v >>= 1;
	}
	return i;
    }

    static int get_lsb(int v) {
	int j = 32;
	int i = 0;
	while (j--) {
	    if (v & 1) return i;
	    v >>= 1;
	    i++;
	}
	return -1;
    }

    LPDIRECTDRAW7		_dd;        // DirectDraw object
    LPDIRECTDRAWSURFACE7	_dd_prisurf;   // DirectDraw primary surface
    LPDIRECTDRAWSURFACE7	_dd_backsurf;  // DirectDraw back surface

    // these are values to shift the color (right) to get them from 8 bits into
    // the proper size to display
    WORD			_RedColorShift;
    WORD			_GreenColorShift;
    WORD			_BlueColorShift;

    // these are values to shift the color (left) to get them into place
    WORD			_RedPosShift;
    WORD			_GreenPosShift;
    WORD			_BluePosShift;

    int				_bpp;
    HWND			_hwnd;

public:
    DirectDrawWin() {
	_dd = NULL;
	_dd_prisurf = NULL;
	_dd_backsurf = NULL;
    }
    virtual ~DirectDrawWin() {
	if (_dd_prisurf) _dd_prisurf->Release();
	if (_dd_backsurf) _dd_backsurf->Release();
	if (_dd) _dd->Release();
    }

    bool init(HWND hwnd) {
	_hwnd = hwnd;

	DDSURFACEDESC2 ddsd;
	LPDIRECTDRAWCLIPPER clipper;

	// Create the main DirectDraw object
	if (_dd == NULL) {
	    if (DirectDrawCreateEx(NULL, (void**)&_dd, IID_IDirectDraw7, NULL) != DD_OK)
		return false;

	    // using DDSCL_NORMAL means we will coexist with GDI
	    if (_dd->SetCooperativeLevel(hwnd, DDSCL_NORMAL) != DD_OK )
		return false;
	}

	// Create primary surface, non page flipping surface since this is
	// windowed mode
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	HRESULT hr;
	if ((hr=_dd->CreateSurface(&ddsd, &_dd_prisurf, NULL)) != DD_OK) {
	    printf("could not get create primary surface: %x\n", hr);
	    fflush(stdout);
	    return false;
	}

	// Get the pixel format of the primary surface
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_PIXELFORMAT;
	if (_dd_prisurf->GetSurfaceDesc(&ddsd) != DD_OK ) {
	    _dd_prisurf->Release();
	    _dd_prisurf = NULL;
	    printf("could not get surface description\n");
	    fflush(stdout);
	    return false;
	}
	_bpp = ddsd.ddpfPixelFormat.dwRGBBitCount;

	_RedColorShift = 8 - count_bits(ddsd.ddpfPixelFormat.dwRBitMask);
	_GreenColorShift = 8 - count_bits(ddsd.ddpfPixelFormat.dwGBitMask);
	_BlueColorShift = 8 - count_bits(ddsd.ddpfPixelFormat.dwBBitMask);

	_RedPosShift = get_lsb(ddsd.ddpfPixelFormat.dwRBitMask);
	_GreenPosShift = get_lsb(ddsd.ddpfPixelFormat.dwGBitMask);
	_BluePosShift = get_lsb(ddsd.ddpfPixelFormat.dwBBitMask);

/*	printf("rgbcount %d, r(%x) g(%x) b(%x)\n",
	       ddsd.ddpfPixelFormat.dwRGBBitCount,
	       ddsd.ddpfPixelFormat.dwRBitMask,
	       ddsd.ddpfPixelFormat.dwGBitMask,
	       ddsd.ddpfPixelFormat.dwBBitMask);
	printf("color shift r(%d) g(%d) b(%d)\npos shift r(%d) g(%d) b(%d)\n",
	       _RedColorShift, _GreenColorShift, _BlueColorShift,
	       _RedPosShift, _GreenPosShift, _BluePosShift);
	fflush(stdout);
 */

	// Create a clipper to ensure that our drawing stays inside our window
	if (_dd->CreateClipper(0, &clipper, NULL) != DD_OK ) {
	    _dd_prisurf->Release();
	    _dd_prisurf = NULL;
	    printf("could not create clipper\n");
	    fflush(stdout);
	    return false;
	}

	// Setting it to our hwnd gives the clipper the coordinates from our
	// window
	if (clipper->SetHWnd(0, hwnd) != DD_OK) {
	    clipper->Release();
	    _dd_prisurf->Release();
	    _dd_prisurf = NULL;
	    printf("could not set hwnd for clipper\n");
	    fflush(stdout);
	    return false;
	}

	// Attach the clipper to the primary surface
	if (_dd_prisurf->SetClipper(clipper) != DD_OK) {
	    clipper->Release();
	    _dd_prisurf->Release();
	    _dd_prisurf = NULL;
	    printf("could not set hwnd for primary surface\n");
	    fflush(stdout);
	    return false;
	}

	// Create a back surface that we will actually write into, and then to
	// display it, we will blt it to primary surface
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);

	RECT r;
	GetClientRect(_hwnd, &r);
//	printf("DDraw window ClientRect size: %d x %d\n", r.right, r.bottom);

	ddsd.ddpfPixelFormat.dwRGBBitCount,
	ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
	ddsd.dwWidth = r.right;
	ddsd.dwHeight = r.bottom;

	if (_dd->CreateSurface(&ddsd, &_dd_backsurf, NULL) != DD_OK ) {
	    clipper->Release();
	    _dd_prisurf->Release();
	    _dd_prisurf = NULL;
	    printf("could not create back surface\n");
	    fflush(stdout);
	    return false;
	}

/*	{ int w = r.right; int h = r.bottom; int j = 0; int pii; u_char red = 0;
	    u_char *p = GetBuffer(&pii);
	    while (j < h) {
		int i = 0; u_char b = 0;
		while (i < w) {
		    p += set_pixel(p, 0, 0, 0);
		    //p += set_pixel(p, red, 0, b);
		    b++; i++;
		} r++; j++;
	    } ReleaseBuffer();
	}
 */

	// We don't need this anymore
	clipper->Release();

	{
	    int pitch;
	    u_char *buf = GetBuffer(&pitch);
	    memset(buf, 0, pitch * ddsd.dwHeight);
	    ReleaseBuffer();
	    draw();
	}

	return true;
    }

    int set_pixel(u_char *buf, u_char r, u_char g, u_char b) {
	if (_bpp == 32) {
	    u_int32_t pix =
		(r >> _RedColorShift) << _RedPosShift |
		(g >> _GreenColorShift) << _GreenPosShift |
		(b >> _BlueColorShift) << _BluePosShift;

	    *(u_int32_t*)buf = pix;
	    return 4;
	} else if (_bpp == 24) {
	    u_int32_t pix =
		(r >> _RedColorShift) << _RedPosShift |
		(g >> _GreenColorShift) << _GreenPosShift |
		(b >> _BlueColorShift) << _BluePosShift;

	    buf[0] = ((u_char*)&pix)[0];
	    buf[1] = ((u_char*)&pix)[1];
	    buf[2] = ((u_char*)&pix)[2];
	    return 3;

	} else if (_bpp == 16) {
	    u_int16_t pix =
		(r >> _RedColorShift) << _RedPosShift |
		(g >> _GreenColorShift) << _GreenPosShift |
		(b >> _BlueColorShift) << _BluePosShift;

	    *(u_int16_t*)buf = pix;
	    return 2;
	}

	assert(!"invalid pixel depth");
	return 1;
    }

    int get_pixel(u_char *p, u_char *r, u_char *g, u_char *b) {
	if (_bpp == 32) {
	    *r = p[2]; *g = p[1]; *b = p[0]; return 4;
	} else if (_bpp == 24) {
	    *r = p[2]; *g = p[1]; *b = p[0]; return 3;
	} else if (_bpp == 16) {
	    *r = ((((p[1] << 8) | p[0]) >> 10) & 31) << 3;
	    *g = ((((p[1] << 8) | p[0]) >> 5) & 31) << 3;
	    *b = (p[0] & 31) << 3;
	    return 2;
	}
	assert(!"invalid dd:get_pixel depth");
	return 1;
    }

    int get_bits_per_pixel(void) {
	return _bpp;
    }

    // Retrieve the area to write into (this is the back surface)
    u_char *GetBuffer(int *pitch) {
	DDSURFACEDESC2 ddsd;
	ddsd.dwSize = sizeof(ddsd);
	_dd_backsurf->Lock(NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR, NULL);
	*pitch = ddsd.lPitch;
	return (u_char*)ddsd.lpSurface;
    }

    // Let go of back surface, must be done before draw() is called
    void ReleaseBuffer(void) {
	_dd_backsurf->Unlock(NULL);
    }


    // Draws the backsurface by blt'ing it onto the primary surface
    void draw(void) {
	RECT rc;
	GetWindowRect(_hwnd, &rc);
	if (_dd_prisurf) {
	    HRESULT r = _dd_prisurf->Blt(&rc, _dd_backsurf, NULL, 0, NULL);
//	    _dd_prisurf->BltFast(rc.left,rc.top, _dd_backsurf, NULL,
//				 DDBLTFAST_NOCOLORKEY|DDBLTFAST_WAIT);
	    if (r != DD_OK) {
		printf("BLT ERR: %x\n", r);
		fflush(stdout);
	    }
	}
    }
};

#endif
